import logging
from datetime import date, datetime, timedelta

from apscheduler.schedulers.blocking import BlockingScheduler
from apscheduler.triggers.cron import CronTrigger

from slack_scheduler.config import AppConfig, Credentials, resolve_skip_dates
from slack_scheduler.selector import pick_message
from slack_scheduler.sender import send_message
from slack_scheduler.templates import render

log = logging.getLogger(__name__)


def run_daemon(
    config: AppConfig,
    credentials: Credentials,
    dry_run: bool = False,
) -> None:
    scheduler = BlockingScheduler()

    for channel in config.channels:
        for i, schedule in enumerate(channel.schedules):
            skip_dates = resolve_skip_dates(config.skip_dates, schedule.skip_dates)

            scheduler.add_job(
                _fire,
                trigger=CronTrigger.from_crontab(schedule.cron),
                jitter=schedule.jitter_minutes * 60 if schedule.jitter_minutes else None,
                args=[channel.id, channel.name, channel.messages, channel.selection_mode,
                      schedule.skip_weekends, skip_dates, credentials, config.workspace_url,
                      dry_run],
                id=f"{channel.id}_{i}",
                name=f"{channel.name} ({schedule.cron})",
            )
            log.info(
                f"Scheduled: {channel.name} ({schedule.cron})"
                f"{f', up to {schedule.jitter_minutes}min jitter' if schedule.jitter_minutes else ''}"
            )

    job_count = len(scheduler.get_jobs())
    if job_count == 0:
        log.warning("No schedules configured. Nothing to do.")
        return

    log.info(f"Starting scheduler with {job_count} job(s). Press Ctrl-C to stop.")
    scheduler.start()


def _fire(
    channel_id: str,
    channel_name: str,
    messages: list[str],
    selection_mode: str,
    skip_weekends: bool,
    skip_dates: set[date],
    credentials: Credentials,
    workspace_url: str,
    dry_run: bool,
) -> None:
    today = date.today()

    if skip_weekends and today.weekday() >= 5:
        log.info(f"Skipping {channel_name}: weekend")
        return

    if today in skip_dates:
        log.info(f"Skipping {channel_name}: {today} is in skip_dates")
        return

    message = pick_message(channel_id, messages, selection_mode)
    message = render(message, datetime.now())

    result = send_message(
        channel_id=channel_id,
        message=message,
        credentials=credentials,
        workspace_url=workspace_url,
        dry_run=dry_run,
    )

    if result.ok and not dry_run:
        log.info(f"Sent to {channel_name}: {message!r} (ts={result.ts})")
    elif not result.ok:
        log.error(f"Failed to send to {channel_name}: {result.error_code}")


def print_upcoming(config: AppConfig, count: int = 5) -> None:
    if not config.channels:
        print("No schedules configured.")
        return

    print(f"\nUpcoming scheduled messages (next {count} per schedule):\n")

    for channel in config.channels:
        for schedule in channel.schedules:
            trigger = CronTrigger.from_crontab(schedule.cron)
            skip_dates = resolve_skip_dates(config.skip_dates, schedule.skip_dates)
            label = f"{channel.name} ({schedule.cron})"
            if schedule.jitter_minutes:
                label += f" up to {schedule.jitter_minutes}min jitter"

            print(f"  {label}")

            now = datetime.now(tz=trigger.timezone)
            upcoming = []
            cursor = now
            iterations = 0
            while len(upcoming) < count and iterations < 1000:
                iterations += 1
                next_time = trigger.get_next_fire_time(cursor, cursor)
                if next_time is None:
                    break
                cursor = next_time + timedelta(seconds=1)
                if schedule.skip_weekends and next_time.date().weekday() >= 5:
                    continue
                if next_time.date() in skip_dates:
                    continue
                upcoming.append(next_time)

            for t in upcoming:
                print(f"    - {t.strftime('%Y-%m-%d %H:%M:%S')}")
            print()



"""Tests for slack_scheduler.sender"""
import pytest
import requests as req_lib
from unittest.mock import call, patch

from slack_scheduler.auth import TokenExpiredError
from slack_scheduler.sender import SendResult, SlackAPIError, _post, send_message
from tests.conftest import make_rate_limited_response, make_slack_response

WORKSPACE = "https://test.slack.com"


# --- SlackAPIError -----------------------------------------------------------

def test_slack_api_error_stores_error_code():
    err = SlackAPIError("channel_not_found")
    assert err.error_code == "channel_not_found"
    assert "channel_not_found" in str(err)


# --- SendResult defaults ----------------------------------------------------

def test_send_result_defaults():
    r = SendResult(ok=True, channel_id="C1", message="hi")
    assert r.ts is None
    assert r.error_code is None


# --- send_message: dry_run ---------------------------------------------------

class TestDryRun:
    def test_returns_ok_without_posting(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            result = send_message("C1", "hello", credentials_obj, WORKSPACE, dry_run=True)
        mock_post.assert_not_called()
        assert result.ok is True
        assert result.channel_id == "C1"
        assert result.message == "hello"


# --- send_message: success ---------------------------------------------------

class TestSuccess:
    def test_returns_ok_with_ts(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True, {"ts": "123.456"})
            result = send_message("C1", "hi", credentials_obj, WORKSPACE)
        assert result.ok is True
        assert result.ts == "123.456"

    def test_single_attempt_on_first_success(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True, {"ts": "1"})
            send_message("C1", "hi", credentials_obj, WORKSPACE)
        assert mock_post.call_count == 1


# --- send_message: retry on network error -----------------------------------

class TestRetry:
    def test_retries_then_succeeds(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post, \
             patch("slack_scheduler.sender.time.sleep"):
            mock_post.side_effect = [
                req_lib.ConnectionError("timeout"),
                make_slack_response(True, {"ts": "1"}),
            ]
            result = send_message("C1", "hi", credentials_obj, WORKSPACE, max_attempts=3)
        assert result.ok is True
        assert mock_post.call_count == 2

    def test_raises_after_all_attempts_exhausted(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post, \
             patch("slack_scheduler.sender.time.sleep"):
            mock_post.side_effect = req_lib.ConnectionError("down")
            with pytest.raises(req_lib.ConnectionError):
                send_message("C1", "hi", credentials_obj, WORKSPACE, max_attempts=3)
        assert mock_post.call_count == 3

    def test_exponential_backoff(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post, \
             patch("slack_scheduler.sender.time.sleep") as mock_sleep:
            mock_post.side_effect = req_lib.ConnectionError("e")
            with pytest.raises(req_lib.ConnectionError):
                send_message("C1", "hi", credentials_obj, WORKSPACE, max_attempts=3)
        # backoff = 2**attempt where attempt is already incremented:
        # attempt=1 fail -> sleep(2**1=2), attempt=2 fail -> sleep(2**2=4), attempt=3 -> raise
        assert mock_sleep.call_args_list == [call(2), call(4)]


# --- send_message: rate limiting --------------------------------------------

class TestRateLimit:
    def test_retries_on_rate_limit_then_succeeds(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post, \
             patch("slack_scheduler.sender.time.sleep"):
            mock_post.side_effect = [
                make_rate_limited_response(retry_after=1),
                make_slack_response(True, {"ts": "1"}),
            ]
            result = send_message("C1", "hi", credentials_obj, WORKSPACE,
                                  max_rate_limit_retries=5)
        assert result.ok is True

    def test_sleeps_for_retry_after_header_value(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post, \
             patch("slack_scheduler.sender.time.sleep") as mock_sleep:
            mock_post.side_effect = [
                make_rate_limited_response(retry_after=30),
                make_slack_response(True, {"ts": "1"}),
            ]
            send_message("C1", "hi", credentials_obj, WORKSPACE,
                         max_rate_limit_retries=5)
        mock_sleep.assert_called_once_with(30)

    def test_gives_up_after_max_rate_limit_retries(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post, \
             patch("slack_scheduler.sender.time.sleep"):
            mock_post.return_value = make_rate_limited_response()
            result = send_message("C1", "hi", credentials_obj, WORKSPACE,
                                  max_rate_limit_retries=2)
        assert result.ok is False
        assert result.error_code == "ratelimited"
        # Should attempt max_rate_limit_retries + 1 times (retries 1,2 then give up on 3)
        assert mock_post.call_count == 3

    def test_rate_limits_do_not_consume_attempt_budget(self, credentials_obj):
        """3 rate limits followed by success should work with max_attempts=1."""
        with patch("slack_scheduler.sender.requests.post") as mock_post, \
             patch("slack_scheduler.sender.time.sleep"):
            mock_post.side_effect = [
                make_rate_limited_response(),
                make_rate_limited_response(),
                make_rate_limited_response(),
                make_slack_response(True, {"ts": "ok"}),
            ]
            result = send_message("C1", "hi", credentials_obj, WORKSPACE,
                                  max_attempts=1, max_rate_limit_retries=5)
        assert result.ok is True


# --- send_message: API errors -----------------------------------------------

class TestAPIErrors:
    def test_invalid_auth_raises_token_expired(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(False, {"error": "invalid_auth"})
            with pytest.raises(TokenExpiredError):
                send_message("C1", "hi", credentials_obj, WORKSPACE)

    @pytest.mark.parametrize("error_code", [
        "channel_not_found",
        "not_in_channel",
        "msg_too_long",
    ])
    def test_other_errors_raise_slack_api_error(self, credentials_obj, error_code):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(False, {"error": error_code})
            with pytest.raises(SlackAPIError) as exc_info:
                send_message("C1", "hi", credentials_obj, WORKSPACE)
        assert exc_info.value.error_code == error_code

    def test_missing_error_key_defaults_to_unknown(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(False)
            with pytest.raises(SlackAPIError) as exc_info:
                send_message("C1", "hi", credentials_obj, WORKSPACE)
        assert exc_info.value.error_code == "unknown"


# --- send_message: JSON decode failure --------------------------------------

class TestJSONDecodeFailure:
    def test_non_json_response_falls_back_to_raise_for_status(self, credentials_obj):
        from unittest.mock import MagicMock
        response = MagicMock()
        response.json.side_effect = ValueError("No JSON")
        response.raise_for_status.side_effect = req_lib.HTTPError("502 Bad Gateway")
        with patch("slack_scheduler.sender.requests.post", return_value=response), \
             patch("slack_scheduler.sender.time.sleep"):
            with pytest.raises(req_lib.HTTPError):
                send_message("C1", "hi", credentials_obj, WORKSPACE, max_attempts=1)


# --- _post: payload shape ---------------------------------------------------

class TestPost:
    def test_url_construction(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True)
            _post("C1", "hi", credentials_obj, "https://test.slack.com")
        assert mock_post.call_args[0][0] == "https://test.slack.com/api/chat.postMessage"

    def test_strips_trailing_slash(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True)
            _post("C1", "hi", credentials_obj, "https://test.slack.com/")
        url = mock_post.call_args[0][0]
        assert "//api" not in url

    def test_sends_bearer_token(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True)
            _post("C1", "hi", credentials_obj, WORKSPACE)
        headers = mock_post.call_args.kwargs["headers"]
        assert headers["Authorization"] == f"Bearer {credentials_obj.xoxc_token}"

    def test_sends_d_cookie(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True)
            _post("C1", "hi", credentials_obj, WORKSPACE)
        assert mock_post.call_args.kwargs["cookies"]["d"] == credentials_obj.d_cookie

    def test_sends_channel_and_text_in_body(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True)
            _post("C1", "hello", credentials_obj, WORKSPACE)
        body = mock_post.call_args.kwargs["json"]
        assert body == {"channel": "C1", "text": "hello"}

    def test_uses_30s_timeout(self, credentials_obj):
        with patch("slack_scheduler.sender.requests.post") as mock_post:
            mock_post.return_value = make_slack_response(True)
            _post("C1", "hi", credentials_obj, WORKSPACE)
        assert mock_post.call_args.kwargs["timeout"] == 30



"""Tests for slack_scheduler.config"""
import textwrap
from datetime import date
from pathlib import Path

import pytest

from slack_scheduler.config import (
    AppConfig,
    ScheduleConfig,
    _validate_skip_dates,
    load_config,
    load_credentials,
    resolve_skip_dates,
)


# --- _validate_skip_dates ---------------------------------------------------

class TestValidateSkipDates:
    def test_valid_dates_returned_unchanged(self):
        dates = ["2026-12-25", "2026-01-01"]
        assert _validate_skip_dates(dates, "ctx") == dates

    def test_empty_list_is_valid(self):
        assert _validate_skip_dates([], "ctx") == []

    @pytest.mark.parametrize("bad_date", [
        "25-12-2026",
        "2026/12/25",
        "not-a-date",
        "2026-13-01",
        "",
    ])
    def test_invalid_format_raises_value_error(self, bad_date):
        with pytest.raises(ValueError, match="expected YYYY-MM-DD"):
            _validate_skip_dates([bad_date], "ctx")

    def test_error_message_contains_context(self):
        with pytest.raises(ValueError, match="my context"):
            _validate_skip_dates(["bad"], "my context")


# --- resolve_skip_dates ------------------------------------------------------

class TestResolveSkipDates:
    def test_combines_global_and_schedule_dates(self):
        result = resolve_skip_dates(["2026-12-25"], ["2026-07-04"])
        assert date(2026, 12, 25) in result
        assert date(2026, 7, 4) in result

    def test_deduplicates_overlapping_dates(self):
        result = resolve_skip_dates(["2026-12-25"], ["2026-12-25"])
        assert len(result) == 1

    def test_empty_inputs_return_empty_set(self):
        assert resolve_skip_dates([], []) == set()

    def test_returns_set_of_date_objects(self):
        result = resolve_skip_dates(["2026-01-01"], [])
        assert all(isinstance(d, date) for d in result)


# --- load_credentials --------------------------------------------------------

class TestLoadCredentials:
    def test_loads_valid_env_file(self, mock_creds_env):
        creds = load_credentials(mock_creds_env)
        assert creds.xoxc_token == "xoxc-test-token-abc123"
        assert creds.d_cookie == "xoxd-test-cookie-def456"

    def test_missing_file_raises_file_not_found(self, tmp_path):
        with pytest.raises(FileNotFoundError, match="not found"):
            load_credentials(tmp_path / "missing.env")

    def test_empty_values_raise_value_error(self, empty_creds_env):
        with pytest.raises(ValueError, match="SLACK_XOXC_TOKEN"):
            load_credentials(empty_creds_env)

    def test_missing_token_key_raises_value_error(self, tmp_path):
        f = tmp_path / "creds.env"
        f.write_text("SLACK_D_COOKIE=xoxd-abc\n")
        with pytest.raises(ValueError):
            load_credentials(f)

    def test_missing_cookie_key_raises_value_error(self, tmp_path):
        f = tmp_path / "creds.env"
        f.write_text("SLACK_XOXC_TOKEN=xoxc-abc\n")
        with pytest.raises(ValueError):
            load_credentials(f)

    def test_empty_file_raises_value_error(self, tmp_path):
        f = tmp_path / "creds.env"
        f.write_text("")
        with pytest.raises(ValueError):
            load_credentials(f)


# --- load_config: happy path -------------------------------------------------

class TestLoadConfigHappyPath:
    def test_loads_minimal_config(self, minimal_config_file):
        cfg = load_config(minimal_config_file)
        assert cfg.workspace_url == "https://test.slack.com"
        assert len(cfg.channels) == 1

    def test_loads_full_config(self, full_config_file):
        cfg = load_config(full_config_file)
        assert len(cfg.channels) == 2
        assert cfg.default_selection_mode == "cycle"

    def test_channel_fields_parsed(self, minimal_config_file):
        ch = load_config(minimal_config_file).channels[0]
        assert ch.id == "C111"
        assert ch.name == "general"
        assert ch.messages == ["Hello!"]

    def test_schedule_fields_parsed(self, minimal_config_file):
        s = load_config(minimal_config_file).channels[0].schedules[0]
        assert isinstance(s, ScheduleConfig)
        assert s.cron == "0 9 * * 1-5"
        assert s.jitter_minutes == 0
        assert s.skip_weekends is False

    def test_global_skip_dates_parsed(self, full_config_file):
        cfg = load_config(full_config_file)
        assert "2026-12-25" in cfg.skip_dates

    def test_schedule_skip_dates_parsed(self, full_config_file):
        cfg = load_config(full_config_file)
        assert "2026-07-04" in cfg.channels[0].schedules[0].skip_dates

    def test_channel_name_defaults_to_id(self, tmp_path):
        p = tmp_path / "config.yaml"
        p.write_text(textwrap.dedent("""\
            workspace_url: "https://test.slack.com"
            channels:
              - id: "C999"
                messages: ["hi"]
                schedules:
                  - cron: "0 9 * * *"
        """))
        assert load_config(p).channels[0].name == "C999"

    def test_selection_mode_inherits_from_default(self, tmp_path):
        p = tmp_path / "config.yaml"
        p.write_text(textwrap.dedent("""\
            workspace_url: "https://test.slack.com"
            default_selection_mode: "cycle"
            channels:
              - id: "C111"
                messages: ["hi"]
                schedules:
                  - cron: "0 9 * * *"
        """))
        assert load_config(p).channels[0].selection_mode == "cycle"

    def test_no_channels_key_returns_empty_list(self, tmp_path):
        p = tmp_path / "config.yaml"
        p.write_text('workspace_url: "https://test.slack.com"\n')
        assert load_config(p).channels == []

    def test_default_selection_mode_defaults_to_random(self, minimal_config_file):
        assert load_config(minimal_config_file).default_selection_mode == "random"


# --- load_config: validation errors -----------------------------------------

class TestLoadConfigValidation:
    def _write(self, tmp_path, text):
        p = tmp_path / "config.yaml"
        p.write_text(textwrap.dedent(text))
        return p

    def test_empty_file_raises(self, tmp_path):
        p = tmp_path / "config.yaml"
        p.write_text("")
        with pytest.raises(ValueError, match="empty"):
            load_config(p)

    def test_missing_workspace_url_raises(self, tmp_path):
        p = self._write(tmp_path, "channels: []\n")
        with pytest.raises(ValueError, match="workspace_url"):
            load_config(p)

    def test_invalid_default_selection_mode_raises(self, tmp_path):
        p = self._write(tmp_path, """\
            workspace_url: "https://test.slack.com"
            default_selection_mode: "weekly"
            channels: []
        """)
        with pytest.raises(ValueError, match="default_selection_mode"):
            load_config(p)

    def test_channel_missing_id_raises(self, tmp_path):
        p = self._write(tmp_path, """\
            workspace_url: "https://test.slack.com"
            channels:
              - name: "oops"
                messages: ["hi"]
                schedules:
                  - cron: "0 9 * * *"
        """)
        with pytest.raises(ValueError, match="missing required field 'id'"):
            load_config(p)

    def test_schedule_missing_cron_raises(self, tmp_path):
        p = self._write(tmp_path, """\
            workspace_url: "https://test.slack.com"
            channels:
              - id: "C111"
                messages: ["hi"]
                schedules:
                  - skip_weekends: true
        """)
        with pytest.raises(ValueError, match="missing required field 'cron'"):
            load_config(p)

    def test_invalid_channel_selection_mode_raises(self, tmp_path):
        p = self._write(tmp_path, """\
            workspace_url: "https://test.slack.com"
            channels:
              - id: "C111"
                selection_mode: "bogus"
                messages: ["hi"]
                schedules:
                  - cron: "0 9 * * *"
        """)
        with pytest.raises(ValueError, match="invalid selection_mode"):
            load_config(p)

    def test_invalid_global_skip_date_raises(self, tmp_path):
        p = self._write(tmp_path, """\
            workspace_url: "https://test.slack.com"
            skip_dates:
              - "25/12/2026"
            channels: []
        """)
        with pytest.raises(ValueError, match="global skip_dates"):
            load_config(p)

    def test_invalid_schedule_skip_date_raises(self, tmp_path):
        p = self._write(tmp_path, """\
            workspace_url: "https://test.slack.com"
            channels:
              - id: "C111"
                messages: ["hi"]
                schedules:
                  - cron: "0 9 * * *"
                    skip_dates:
                      - "not-a-date"
        """)
        with pytest.raises(ValueError, match="skip_dates"):
            load_config(p)


# --- dataclass default isolation --------------------------------------------

class TestDataclassDefaults:
    def test_schedule_config_skip_dates_not_shared(self):
        a = ScheduleConfig(cron="0 9 * * *")
        b = ScheduleConfig(cron="0 9 * * *")
        a.skip_dates.append("2026-01-01")
        assert b.skip_dates == []

    def test_app_config_skip_dates_not_shared(self):
        a = AppConfig(workspace_url="x", channels=[])
        b = AppConfig(workspace_url="y", channels=[])
        a.skip_dates.append("2026-01-01")
        assert b.skip_dates == []



"""Tests for slack_scheduler.templates"""
import logging
from datetime import datetime

import pytest

from slack_scheduler.templates import _SafeDict, render

FIXED_DT = datetime(2026, 3, 3, 9, 15, 0)  # Tuesday


class TestSafeDict:
    def test_returns_value_for_known_key(self):
        d = _SafeDict(foo="bar")
        assert d["foo"] == "bar"

    def test_returns_placeholder_for_missing_key(self):
        d = _SafeDict()
        assert d["missing"] == "{missing}"


class TestRender:
    def test_renders_date_variable(self):
        assert render("Today is {date}", now=FIXED_DT) == "Today is 2026-03-03"

    def test_renders_day_of_week_variable(self):
        assert render("Happy {day_of_week}!", now=FIXED_DT) == "Happy Tuesday!"

    def test_renders_time_variable(self):
        assert render("It is {time}", now=FIXED_DT) == "It is 09:15"

    def test_renders_multiple_variables(self):
        result = render("{day_of_week} {date} at {time}", now=FIXED_DT)
        assert result == "Tuesday 2026-03-03 at 09:15"

    def test_unknown_placeholder_preserved(self):
        assert render("Hello {name}!", now=FIXED_DT) == "Hello {name}!"

    def test_mixed_known_and_unknown_placeholders(self):
        result = render("{date} - {unknown}", now=FIXED_DT)
        assert result == "2026-03-03 - {unknown}"

    def test_no_placeholders_unchanged(self):
        assert render("Plain text", now=FIXED_DT) == "Plain text"

    def test_empty_string_returns_empty(self):
        assert render("", now=FIXED_DT) == ""

    def test_uses_current_time_when_now_is_none(self):
        result = render("{date}")
        assert len(result) == 10  # YYYY-MM-DD format

    def test_malformed_format_returns_template_as_is(self):
        bad = "Hello {name!badconv}"
        assert render(bad, now=FIXED_DT) == bad

    def test_malformed_format_logs_warning(self, caplog):
        with caplog.at_level(logging.WARNING, logger="slack_scheduler.templates"):
            render("Hello {name!badconv}", now=FIXED_DT)
        assert "Failed to render template" in caplog.text

    @pytest.mark.parametrize("template,expected_substr", [
        ("Morning! Today is {date}.", "2026-03-03"),
        ("{day_of_week} standup", "Tuesday"),
        ("Meeting at {time}", "09:15"),
    ])
    def test_parametrized_substitution(self, template, expected_substr):
        assert expected_substr in render(template, now=FIXED_DT)



import logging
from datetime import datetime

log = logging.getLogger(__name__)


class _SafeDict(dict):
    """Returns the key placeholder for missing keys instead of raising."""

    def __missing__(self, key):
        return f"{{{key}}}"


def render(template: str, now: datetime | None = None) -> str:
    now = now or datetime.now()
    variables = _SafeDict(
        date=now.strftime("%Y-%m-%d"),
        day_of_week=now.strftime("%A"),
        time=now.strftime("%H:%M"),
    )
    try:
        return template.format_map(variables)
    except ValueError:
        log.warning(f"Failed to render template (bad format syntax), sending as-is: {template!r}")
        return template



"""Tests for slack_scheduler.scheduler"""
import re
from datetime import date, datetime
from unittest.mock import MagicMock, patch

import pytest

from slack_scheduler.config import AppConfig, ChannelConfig, ScheduleConfig
from slack_scheduler.scheduler import _fire, print_upcoming, run_daemon
from slack_scheduler.sender import SendResult


# --- run_daemon: job registration -------------------------------------------

class TestRunDaemon:
    def test_registers_one_job_per_schedule(self, app_config, credentials_obj):
        with patch("slack_scheduler.scheduler.BlockingScheduler") as MockSched:
            instance = MockSched.return_value
            instance.get_jobs.return_value = [MagicMock()]
            run_daemon(app_config, credentials_obj, dry_run=True)
        instance.add_job.assert_called_once()

    def test_registers_multiple_schedules(self, credentials_obj):
        channel = ChannelConfig(
            id="C1", name="ch", messages=["hi"],
            schedules=[
                ScheduleConfig(cron="0 9 * * 1-5"),
                ScheduleConfig(cron="0 14 * * 3"),
            ],
        )
        config = AppConfig(workspace_url="https://test.slack.com", channels=[channel])
        with patch("slack_scheduler.scheduler.BlockingScheduler") as MockSched:
            instance = MockSched.return_value
            instance.get_jobs.return_value = [MagicMock(), MagicMock()]
            run_daemon(config, credentials_obj)
        assert instance.add_job.call_count == 2

    def test_no_schedules_does_not_start(self, credentials_obj):
        config = AppConfig(workspace_url="https://test.slack.com", channels=[])
        with patch("slack_scheduler.scheduler.BlockingScheduler") as MockSched:
            instance = MockSched.return_value
            instance.get_jobs.return_value = []
            run_daemon(config, credentials_obj)
        instance.start.assert_not_called()

    def test_job_id_format(self, app_config, credentials_obj):
        with patch("slack_scheduler.scheduler.BlockingScheduler") as MockSched:
            instance = MockSched.return_value
            instance.get_jobs.return_value = [MagicMock()]
            run_daemon(app_config, credentials_obj)
        assert instance.add_job.call_args.kwargs["id"] == "C111_0"

    def test_jitter_seconds_passed(self, credentials_obj):
        channel = ChannelConfig(
            id="C1", name="ch", messages=["hi"],
            schedules=[ScheduleConfig(cron="0 9 * * *", jitter_minutes=15)],
        )
        config = AppConfig(workspace_url="https://test.slack.com", channels=[channel])
        with patch("slack_scheduler.scheduler.BlockingScheduler") as MockSched:
            instance = MockSched.return_value
            instance.get_jobs.return_value = [MagicMock()]
            run_daemon(config, credentials_obj)
        assert instance.add_job.call_args.kwargs["jitter"] == 15 * 60

    def test_zero_jitter_passes_none(self, app_config, credentials_obj):
        with patch("slack_scheduler.scheduler.BlockingScheduler") as MockSched:
            instance = MockSched.return_value
            instance.get_jobs.return_value = [MagicMock()]
            run_daemon(app_config, credentials_obj)
        assert instance.add_job.call_args.kwargs["jitter"] is None


# --- _fire: skip logic -------------------------------------------------------

CREDS = MagicMock()


def _make_fire_kwargs(**overrides):
    base = dict(
        channel_id="C1", channel_name="general", messages=["hi"],
        selection_mode="random", skip_weekends=False, skip_dates=set(),
        credentials=CREDS, workspace_url="https://test.slack.com", dry_run=False,
    )
    base.update(overrides)
    return base


class TestFire:
    def test_sends_on_normal_weekday(self):
        monday = date(2026, 3, 2)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.pick_message", return_value="hi"), \
             patch("slack_scheduler.scheduler.render", return_value="hi"), \
             patch("slack_scheduler.scheduler.send_message",
                   return_value=SendResult(ok=True, channel_id="C1", message="hi", ts="1")) as mock_send:
            mock_date.today.return_value = monday
            _fire(**_make_fire_kwargs())
        mock_send.assert_called_once()

    def test_skips_saturday_when_skip_weekends(self):
        saturday = date(2026, 3, 7)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.send_message") as mock_send:
            mock_date.today.return_value = saturday
            _fire(**_make_fire_kwargs(skip_weekends=True))
        mock_send.assert_not_called()

    def test_skips_sunday_when_skip_weekends(self):
        sunday = date(2026, 3, 8)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.send_message") as mock_send:
            mock_date.today.return_value = sunday
            _fire(**_make_fire_kwargs(skip_weekends=True))
        mock_send.assert_not_called()

    def test_sends_on_weekend_when_skip_weekends_false(self):
        saturday = date(2026, 3, 7)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.pick_message", return_value="hi"), \
             patch("slack_scheduler.scheduler.render", return_value="hi"), \
             patch("slack_scheduler.scheduler.send_message",
                   return_value=SendResult(ok=True, channel_id="C1", message="hi")) as mock_send:
            mock_date.today.return_value = saturday
            _fire(**_make_fire_kwargs(skip_weekends=False))
        mock_send.assert_called_once()

    def test_skips_when_today_in_skip_dates(self):
        today = date(2026, 12, 25)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.send_message") as mock_send:
            mock_date.today.return_value = today
            _fire(**_make_fire_kwargs(skip_dates={today}))
        mock_send.assert_not_called()

    def test_sends_when_today_not_in_skip_dates(self):
        today = date(2026, 3, 3)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.pick_message", return_value="hi"), \
             patch("slack_scheduler.scheduler.render", return_value="hi"), \
             patch("slack_scheduler.scheduler.send_message",
                   return_value=SendResult(ok=True, channel_id="C1", message="hi")) as mock_send:
            mock_date.today.return_value = today
            _fire(**_make_fire_kwargs(skip_dates={date(2026, 12, 25)}))
        mock_send.assert_called_once()

    def test_dry_run_forwarded_to_send_message(self):
        monday = date(2026, 3, 2)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.pick_message", return_value="hi"), \
             patch("slack_scheduler.scheduler.render", return_value="hi"), \
             patch("slack_scheduler.scheduler.send_message",
                   return_value=SendResult(ok=True, channel_id="C1", message="hi")) as mock_send:
            mock_date.today.return_value = monday
            _fire(**_make_fire_kwargs(dry_run=True))
        assert mock_send.call_args.kwargs["dry_run"] is True

    def test_logs_error_on_failed_send(self):
        monday = date(2026, 3, 2)
        with patch("slack_scheduler.scheduler.date", side_effect=date, today=MagicMock()) as mock_date, \
             patch("slack_scheduler.scheduler.pick_message", return_value="hi"), \
             patch("slack_scheduler.scheduler.render", return_value="hi"), \
             patch("slack_scheduler.scheduler.send_message",
                   return_value=SendResult(ok=False, channel_id="C1", message="hi",
                                           error_code="channel_not_found")):
            mock_date.today.return_value = monday
            # Should not raise, just log
            _fire(**_make_fire_kwargs())


# --- print_upcoming ----------------------------------------------------------

class TestPrintUpcoming:
    def test_empty_config_prints_no_schedules(self, capsys):
        config = AppConfig(workspace_url="https://test.slack.com", channels=[])
        print_upcoming(config)
        assert "No schedules configured" in capsys.readouterr().out

    def test_prints_channel_name_and_cron(self, app_config, capsys):
        print_upcoming(app_config, count=1)
        out = capsys.readouterr().out
        assert "general" in out
        assert "0 9 * * 1-5" in out

    def test_prints_expected_count(self, app_config, capsys):
        print_upcoming(app_config, count=3)
        out = capsys.readouterr().out
        time_lines = [l for l in out.splitlines() if l.strip().startswith("- 20")]
        assert len(time_lines) == 3

    def test_skips_weekends(self, capsys):
        channel = ChannelConfig(
            id="C1", name="ch", messages=["hi"],
            schedules=[ScheduleConfig(cron="0 9 * * *", skip_weekends=True)],
        )
        config = AppConfig(workspace_url="https://test.slack.com", channels=[channel])
        print_upcoming(config, count=7)
        out = capsys.readouterr().out
        date_strs = re.findall(r"\d{4}-\d{2}-\d{2}", out)
        for ds in date_strs:
            dt = datetime.strptime(ds, "%Y-%m-%d")
            assert dt.weekday() < 5, f"{ds} is a weekend day"

    def test_skip_dates_excluded(self, capsys):
        channel = ChannelConfig(
            id="C1", name="ch", messages=["hi"],
            schedules=[ScheduleConfig(cron="0 9 * * *")],
        )
        skip_date = "2026-03-04"
        config = AppConfig(
            workspace_url="https://test.slack.com",
            channels=[channel], skip_dates=[skip_date],
        )
        print_upcoming(config, count=10)
        assert skip_date not in capsys.readouterr().out

    def test_jitter_label_shown(self, capsys):
        channel = ChannelConfig(
            id="C1", name="ch", messages=["hi"],
            schedules=[ScheduleConfig(cron="0 9 * * *", jitter_minutes=10)],
        )
        config = AppConfig(workspace_url="https://test.slack.com", channels=[channel])
        print_upcoming(config, count=1)
        assert "10min jitter" in capsys.readouterr().out

    def test_no_jitter_label_when_zero(self, app_config, capsys):
        print_upcoming(app_config, count=1)
        assert "jitter" not in capsys.readouterr().out



from pathlib import Path

from platformdirs import user_config_dir, user_data_dir

APP_NAME = "slack-scheduler"


def config_dir() -> Path:
    return Path(user_config_dir(APP_NAME))


def data_dir() -> Path:
    return Path(user_data_dir(APP_NAME))


def log_dir() -> Path:
    return data_dir() / "logs"


def state_dir() -> Path:
    return data_dir() / "state"



"""Tests for slack_scheduler.logger"""
import logging

import pytest

from slack_scheduler.logger import setup_logging


@pytest.fixture(autouse=True)
def reset_logger():
    """Ensure the slack_scheduler logger is clean before and after each test."""
    logger = logging.getLogger("slack_scheduler")
    yield
    logger.handlers.clear()


class TestSetupLogging:
    def test_creates_log_file(self, tmp_path):
        setup_logging(log_dir=tmp_path)
        assert (tmp_path / "slack_scheduler.log").exists()

    def test_attaches_exactly_two_handlers(self, tmp_path):
        setup_logging(log_dir=tmp_path)
        logger = logging.getLogger("slack_scheduler")
        assert len(logger.handlers) == 2

    def test_handlers_are_file_and_stream(self, tmp_path):
        setup_logging(log_dir=tmp_path)
        handler_types = {type(h) for h in logging.getLogger("slack_scheduler").handlers}
        assert logging.FileHandler in handler_types
        assert logging.StreamHandler in handler_types

    def test_verbose_false_console_level_info(self, tmp_path):
        setup_logging(verbose=False, log_dir=tmp_path)
        stream = [h for h in logging.getLogger("slack_scheduler").handlers
                  if type(h) is logging.StreamHandler]
        assert stream[0].level == logging.INFO

    def test_verbose_true_console_level_debug(self, tmp_path):
        setup_logging(verbose=True, log_dir=tmp_path)
        stream = [h for h in logging.getLogger("slack_scheduler").handlers
                  if type(h) is logging.StreamHandler]
        assert stream[0].level == logging.DEBUG

    def test_file_handler_always_debug(self, tmp_path):
        setup_logging(verbose=False, log_dir=tmp_path)
        file_h = [h for h in logging.getLogger("slack_scheduler").handlers
                  if isinstance(h, logging.FileHandler)]
        assert file_h[0].level == logging.DEBUG

    def test_repeated_calls_no_duplicate_handlers(self, tmp_path):
        setup_logging(log_dir=tmp_path)
        setup_logging(log_dir=tmp_path)
        assert len(logging.getLogger("slack_scheduler").handlers) == 2

    def test_creates_log_dir_if_missing(self, tmp_path):
        nested = tmp_path / "deep" / "logs"
        setup_logging(log_dir=nested)
        assert nested.is_dir()

    def test_root_logger_level_is_debug(self, tmp_path):
        setup_logging(log_dir=tmp_path)
        assert logging.getLogger("slack_scheduler").level == logging.DEBUG



"""Tests for slack_scheduler.auth"""
import pytest
import requests as req_lib
from unittest.mock import MagicMock, patch

from slack_scheduler.auth import (
    TokenExpiredError,
    TokenInvalidError,
    validate_credentials,
)

WORKSPACE = "https://test.slack.com"


def _mock_auth_response(ok: bool, extra: dict | None = None) -> MagicMock:
    r = MagicMock()
    payload = {"ok": ok}
    if extra:
        payload.update(extra)
    r.json.return_value = payload
    return r


class TestValidateCredentials:
    def test_success_does_not_raise(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(True, {"user": "alice", "team": "ACME"})
            validate_credentials(credentials_obj, WORKSPACE)

    def test_posts_to_auth_test_endpoint(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(True)
            validate_credentials(credentials_obj, WORKSPACE)
        assert mock_post.call_args[0][0] == "https://test.slack.com/api/auth.test"

    def test_strips_trailing_slash(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(True)
            validate_credentials(credentials_obj, "https://test.slack.com/")
        url = mock_post.call_args[0][0]
        assert "//api" not in url

    def test_invalid_auth_raises_token_expired(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(False, {"error": "invalid_auth"})
            with pytest.raises(TokenExpiredError):
                validate_credentials(credentials_obj, WORKSPACE)

    def test_other_error_raises_token_invalid(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(False, {"error": "account_inactive"})
            with pytest.raises(TokenInvalidError, match="account_inactive"):
                validate_credentials(credentials_obj, WORKSPACE)

    def test_missing_error_key_defaults_to_unknown(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(False)
            with pytest.raises(TokenInvalidError, match="unknown"):
                validate_credentials(credentials_obj, WORKSPACE)

    def test_network_error_raises_token_invalid(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post",
                   side_effect=req_lib.ConnectionError("timeout")):
            with pytest.raises(TokenInvalidError, match="Failed to reach Slack API"):
                validate_credentials(credentials_obj, WORKSPACE)

    def test_sends_bearer_token(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(True)
            validate_credentials(credentials_obj, WORKSPACE)
        headers = mock_post.call_args.kwargs["headers"]
        assert headers["Authorization"] == f"Bearer {credentials_obj.xoxc_token}"

    def test_sends_d_cookie(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(True)
            validate_credentials(credentials_obj, WORKSPACE)
        assert mock_post.call_args.kwargs["cookies"]["d"] == credentials_obj.d_cookie

    def test_uses_15s_timeout(self, credentials_obj):
        with patch("slack_scheduler.auth.requests.post") as mock_post:
            mock_post.return_value = _mock_auth_response(True)
            validate_credentials(credentials_obj, WORKSPACE)
        assert mock_post.call_args.kwargs["timeout"] == 15


class TestExceptionClasses:
    def test_token_expired_is_exception(self):
        assert issubclass(TokenExpiredError, Exception)

    def test_token_invalid_is_exception(self):
        assert issubclass(TokenInvalidError, Exception)

    def test_they_are_distinct_classes(self):
        assert TokenExpiredError is not TokenInvalidError



import logging

import requests

from slack_scheduler.config import Credentials

log = logging.getLogger(__name__)


class TokenExpiredError(Exception):
    pass


class TokenInvalidError(Exception):
    pass


def validate_credentials(credentials: Credentials, workspace_url: str) -> None:
    try:
        response = requests.post(
            f"{workspace_url.rstrip('/')}/api/auth.test",
            headers={"Authorization": f"Bearer {credentials.xoxc_token}"},
            cookies={"d": credentials.d_cookie},
            timeout=15,
        )
        data = response.json()
    except requests.RequestException as e:
        raise TokenInvalidError(f"Failed to reach Slack API: {e}")

    if not data.get("ok"):
        error = data.get("error", "unknown")
        if error == "invalid_auth":
            raise TokenExpiredError(
                "Slack credentials have expired. Update your credentials file."
            )
        raise TokenInvalidError(f"Slack auth.test failed: {error}")

    user = data.get("user", "unknown")
    team = data.get("team", "unknown")
    log.info(f"Authenticated as {user} in workspace {team}")



import logging
from pathlib import Path

from slack_scheduler import paths


def setup_logging(verbose: bool = False, log_dir: Path | None = None) -> None:
    log_dir = log_dir or paths.log_dir()
    log_dir.mkdir(parents=True, exist_ok=True)

    log_file = log_dir / "slack_scheduler.log"
    level = logging.DEBUG if verbose else logging.INFO

    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(name)s: %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    file_handler = logging.FileHandler(log_file)
    file_handler.setLevel(logging.DEBUG)
    file_handler.setFormatter(formatter)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(level)
    console_handler.setFormatter(formatter)

    root = logging.getLogger("slack_scheduler")
    root.handlers.clear()
    root.setLevel(logging.DEBUG)
    root.addHandler(file_handler)
    root.addHandler(console_handler)



import argparse
import logging
import platform
import random
import sys
from importlib import resources
from pathlib import Path

from slack_scheduler import paths
from slack_scheduler.auth import TokenExpiredError, TokenInvalidError
from slack_scheduler.logger import setup_logging
from slack_scheduler.sender import SlackAPIError

log = logging.getLogger(__name__)


def main():
    default_config = paths.config_dir() / "config.yaml"
    default_env = paths.data_dir() / "credentials.env"

    parser = argparse.ArgumentParser(
        prog="slack-scheduler",
        description="Schedule and send Slack messages using browser session tokens.",
    )

    # Global flags
    parser.add_argument(
        "--config", type=Path, default=default_config,
        help=f"Path to config.yaml (default: {default_config})",
    )
    parser.add_argument(
        "--env", type=Path, default=default_env,
        help=f"Path to credentials file (default: {default_env})",
    )
    parser.add_argument(
        "--dry-run", action="store_true",
        help="Preview actions without sending anything",
    )
    parser.add_argument(
        "--verbose", action="store_true",
        help="Enable debug-level log output",
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    # init
    subparsers.add_parser(
        "init", help="Create config directories and example config file",
    )

    # send
    send_parser = subparsers.add_parser(
        "send", help="Send a message immediately",
    )
    send_parser.add_argument(
        "--channel", type=str, required=True,
        help="Target channel ID",
    )
    send_parser.add_argument(
        "--message", type=str, nargs="+",
        help="Message text (multiple for random selection)",
    )
    send_parser.add_argument(
        "--workspace", type=str,
        help="Workspace URL (overrides config)",
    )
    send_parser.add_argument(
        "--jitter", type=int, default=0,
        help="Random delay in minutes before sending (e.g. --jitter 15 waits 0-15 min)",
    )
    send_parser.add_argument(
        "--selection-mode", type=str, choices=["random", "cycle"],
        help="Message selection mode (overrides config)",
    )

    # run
    subparsers.add_parser(
        "run", help="Start the daemon scheduler",
    )

    # status
    status_parser = subparsers.add_parser(
        "status", help="Show upcoming scheduled messages",
    )
    status_parser.add_argument(
        "--count", type=int, default=5,
        help="Number of upcoming events to show per schedule (default: 5)",
    )

    # validate
    subparsers.add_parser(
        "validate", help="Check that stored credentials are valid",
    )

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        sys.exit(1)

    setup_logging(verbose=args.verbose)

    try:
        if args.command == "init":
            cmd_init(args)
        elif args.command == "send":
            cmd_send(args)
        elif args.command == "run":
            cmd_run(args)
        elif args.command == "status":
            cmd_status(args)
        elif args.command == "validate":
            cmd_validate(args)
    except TokenExpiredError as e:
        log.error(str(e))
        sys.exit(1)
    except TokenInvalidError as e:
        log.error(str(e))
        sys.exit(1)
    except SlackAPIError as e:
        log.error(str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        log.info("Interrupted. Exiting.")
        sys.exit(0)
    except FileNotFoundError as e:
        log.error(str(e))
        sys.exit(1)
    except ValueError as e:
        log.error(str(e))
        log.debug("ValueError details:", exc_info=True)
        sys.exit(1)
    except Exception as e:
        log.exception(f"Unexpected error: {e}")
        sys.exit(1)


def cmd_init(args):
    config_dir = paths.config_dir()
    data_dir = paths.data_dir()
    log_dir = paths.log_dir()

    for d in [config_dir, data_dir, log_dir]:
        d.mkdir(parents=True, exist_ok=True)

    config_dest = config_dir / "config.yaml"
    if not config_dest.exists():
        try:
            ref = resources.files("slack_scheduler").joinpath("config.example.yaml")
            example_content = ref.read_text()
            config_dest.write_text(example_content)
            print(f"Example config copied to: {config_dest}")
        except (FileNotFoundError, AttributeError):
            print(f"Config directory created: {config_dir}")
            print("Create your config.yaml there to get started.")
    else:
        print(f"Config already exists: {config_dest}")

    env_dest = data_dir / "credentials.env"
    if not env_dest.exists():
        env_dest.write_text(
            "SLACK_XOXC_TOKEN=xoxc-your-token-here\n"
            "SLACK_D_COOKIE=xoxd-your-cookie-here\n"
        )
        if platform.system() != "Windows":
            env_dest.chmod(0o600)
        print(f"Credentials template created: {env_dest}")
    else:
        print(f"Credentials file already exists: {env_dest}")

    print(f"\nPaths:")
    print(f"  Config:      {config_dir}")
    print(f"  Credentials: {env_dest}")
    print(f"  Logs:        {log_dir}")


def cmd_send(args):
    import time
    from datetime import datetime

    from slack_scheduler.auth import validate_credentials
    from slack_scheduler.config import load_config, load_credentials
    from slack_scheduler.sender import send_message
    from slack_scheduler.templates import render

    credentials = load_credentials(args.env)

    config = load_config(args.config) if args.config.exists() else None

    workspace_url = args.workspace
    if not workspace_url and config:
        workspace_url = config.workspace_url
    if not workspace_url:
        log.error("Workspace URL required. Use --workspace or set it in config.yaml.")
        sys.exit(1)

    validate_credentials(credentials, workspace_url)

    if args.message:
        selection_mode = args.selection_mode or (config.default_selection_mode if config else "random")
        if selection_mode == "cycle":
            from slack_scheduler.selector import pick_message
            message = pick_message(args.channel, args.message, "cycle")
        else:
            message = random.choice(args.message)
    elif config:
        channel_cfg = next((c for c in config.channels if c.id == args.channel), None)
        if channel_cfg and channel_cfg.messages:
            from slack_scheduler.selector import pick_message
            mode = args.selection_mode or channel_cfg.selection_mode
            message = pick_message(args.channel, channel_cfg.messages, mode)
        else:
            log.error("No --message provided and no messages in config for this channel.")
            sys.exit(1)
    else:
        log.error("No --message provided and no config file found.")
        sys.exit(1)

    message = render(message, datetime.now())

    if args.jitter > 0:
        delay = random.uniform(0, args.jitter * 60)
        log.info(f"Jitter: waiting {delay:.0f}s before sending")
        time.sleep(delay)

    result = send_message(
        channel_id=args.channel,
        message=message,
        credentials=credentials,
        workspace_url=workspace_url,
        dry_run=args.dry_run,
    )

    if result.ok:
        print(f"Message sent: {message!r}")
    else:
        log.error(f"Failed to send: {result.error_code}")
        sys.exit(1)


def cmd_run(args):
    from slack_scheduler.auth import validate_credentials
    from slack_scheduler.config import load_config, load_credentials
    from slack_scheduler.scheduler import run_daemon

    config = load_config(args.config)
    credentials = load_credentials(args.env)
    validate_credentials(credentials, config.workspace_url)

    run_daemon(config, credentials, dry_run=args.dry_run)


def cmd_status(args):
    from slack_scheduler.config import load_config
    from slack_scheduler.scheduler import print_upcoming

    config = load_config(args.config)
    print_upcoming(config, count=args.count)


def cmd_validate(args):
    from slack_scheduler.auth import validate_credentials
    from slack_scheduler.config import load_config, load_credentials

    credentials = load_credentials(args.env)

    workspace_url = None
    if args.config.exists():
        config = load_config(args.config)
        workspace_url = config.workspace_url

    if not workspace_url:
        log.error("Workspace URL required. Set it in config.yaml.")
        sys.exit(1)

    validate_credentials(credentials, workspace_url)
    print("Credentials are valid.")


if __name__ == "__main__":
    main()



import json
import logging
import os
import random
import tempfile
from pathlib import Path

from slack_scheduler import paths

log = logging.getLogger(__name__)


def pick_message(
    channel_id: str,
    messages: list[str],
    mode: str = "random",
    state_dir: Path | None = None,
) -> str:
    if not messages:
        raise ValueError(f"No messages configured for channel {channel_id}")

    if len(messages) == 1:
        return messages[0]

    if mode == "cycle":
        return _pick_cycle(channel_id, messages, state_dir or paths.state_dir())
    return random.choice(messages)


def _pick_cycle(channel_id: str, messages: list[str], state_dir: Path) -> str:
    state_dir.mkdir(parents=True, exist_ok=True)
    state_file = state_dir / f"{channel_id}.json"

    state = _load_state(state_file)

    try:
        stale = (
            state is None
            or state["index"] >= len(state["shuffled"])
            or sorted(state["shuffled"]) != sorted(messages)
        )
    except (TypeError, KeyError, AttributeError):
        stale = True

    if stale:
        shuffled = messages.copy()
        random.shuffle(shuffled)
        state = {"shuffled": shuffled, "index": 0}

    message = state["shuffled"][state["index"]]
    state["index"] += 1
    _save_state(state_file, state)
    return message


def _load_state(state_file: Path) -> dict | None:
    try:
        with open(state_file) as f:
            return json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        return None


def _save_state(state_file: Path, state: dict) -> None:
    tmp_fd, tmp_path = tempfile.mkstemp(dir=state_file.parent)
    try:
        with os.fdopen(tmp_fd, "w") as f:
            json.dump(state, f)
        os.replace(tmp_path, state_file)
    except Exception:
        os.unlink(tmp_path)
        raise



"""Tests for slack_scheduler.cli

cli.py uses deferred imports inside cmd_* functions, so we must patch at the
source module (e.g., slack_scheduler.config.load_credentials) rather than
slack_scheduler.cli.load_credentials.
"""
import sys
from pathlib import Path
from types import SimpleNamespace
from unittest.mock import MagicMock, patch

import pytest

from slack_scheduler.auth import TokenExpiredError, TokenInvalidError
from slack_scheduler.cli import cmd_init, cmd_send, cmd_status, cmd_validate, main
from slack_scheduler.config import AppConfig, ChannelConfig, ScheduleConfig
from slack_scheduler.sender import SendResult, SlackAPIError


# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------

def make_args(**kwargs) -> SimpleNamespace:
    defaults = dict(
        config=Path("/nonexistent/config.yaml"),
        env=Path("/nonexistent/credentials.env"),
        dry_run=False,
        verbose=False,
    )
    defaults.update(kwargs)
    return SimpleNamespace(**defaults)


# Patch targets for deferred imports used by cmd_* functions
P_LOAD_CREDS = "slack_scheduler.config.load_credentials"
P_LOAD_CONFIG = "slack_scheduler.config.load_config"
P_VALIDATE = "slack_scheduler.auth.validate_credentials"
P_SEND = "slack_scheduler.sender.send_message"
P_RENDER = "slack_scheduler.templates.render"
P_PICK = "slack_scheduler.selector.pick_message"
P_RUN_DAEMON = "slack_scheduler.scheduler.run_daemon"
P_PRINT_UPCOMING = "slack_scheduler.scheduler.print_upcoming"
P_SLEEP = "time.sleep"


# ---------------------------------------------------------------------------
# cmd_init
# ---------------------------------------------------------------------------

class TestCmdInit:
    def test_creates_directories(self, tmp_path):
        with patch("slack_scheduler.cli.paths.config_dir", return_value=tmp_path / "config"), \
             patch("slack_scheduler.cli.paths.data_dir", return_value=tmp_path / "data"), \
             patch("slack_scheduler.cli.paths.log_dir", return_value=tmp_path / "logs"):
            cmd_init(make_args())
        assert (tmp_path / "config").is_dir()
        assert (tmp_path / "data").is_dir()
        assert (tmp_path / "logs").is_dir()

    def test_writes_credentials_template(self, tmp_path, capsys):
        with patch("slack_scheduler.cli.paths.config_dir", return_value=tmp_path / "cfg"), \
             patch("slack_scheduler.cli.paths.data_dir", return_value=tmp_path / "data"), \
             patch("slack_scheduler.cli.paths.log_dir", return_value=tmp_path / "logs"):
            cmd_init(make_args())
        content = (tmp_path / "data" / "credentials.env").read_text()
        assert "SLACK_XOXC_TOKEN" in content
        assert "SLACK_D_COOKIE" in content

    def test_skips_existing_credentials(self, tmp_path, capsys):
        data_dir = tmp_path / "data"
        data_dir.mkdir()
        (data_dir / "credentials.env").write_text("existing")
        with patch("slack_scheduler.cli.paths.config_dir", return_value=tmp_path / "cfg"), \
             patch("slack_scheduler.cli.paths.data_dir", return_value=data_dir), \
             patch("slack_scheduler.cli.paths.log_dir", return_value=tmp_path / "logs"):
            cmd_init(make_args())
        assert (data_dir / "credentials.env").read_text() == "existing"
        assert "already exists" in capsys.readouterr().out

    def test_prints_path_summary(self, tmp_path, capsys):
        with patch("slack_scheduler.cli.paths.config_dir", return_value=tmp_path / "cfg"), \
             patch("slack_scheduler.cli.paths.data_dir", return_value=tmp_path / "data"), \
             patch("slack_scheduler.cli.paths.log_dir", return_value=tmp_path / "logs"):
            cmd_init(make_args())
        assert "Paths:" in capsys.readouterr().out

    def test_chmod_called_on_non_windows(self, tmp_path):
        with patch("slack_scheduler.cli.paths.config_dir", return_value=tmp_path / "cfg"), \
             patch("slack_scheduler.cli.paths.data_dir", return_value=tmp_path / "data"), \
             patch("slack_scheduler.cli.paths.log_dir", return_value=tmp_path / "logs"), \
             patch("slack_scheduler.cli.platform.system", return_value="Linux"):
            cmd_init(make_args())
        assert (tmp_path / "data" / "credentials.env").exists()

    def test_chmod_skipped_on_windows(self, tmp_path):
        with patch("slack_scheduler.cli.paths.config_dir", return_value=tmp_path / "cfg"), \
             patch("slack_scheduler.cli.paths.data_dir", return_value=tmp_path / "data"), \
             patch("slack_scheduler.cli.paths.log_dir", return_value=tmp_path / "logs"), \
             patch("slack_scheduler.cli.platform.system", return_value="Windows"):
            cmd_init(make_args())
        assert (tmp_path / "data" / "credentials.env").exists()


# ---------------------------------------------------------------------------
# cmd_send
# ---------------------------------------------------------------------------

class TestCmdSend:
    def test_sends_cli_message(self, tmp_path, capsys):
        mock_result = SendResult(ok=True, channel_id="C1", message="hello", ts="1")
        args = make_args(
            channel="C1", message=["hello"], workspace="https://test.slack.com",
            jitter=0, selection_mode=None,
            env=tmp_path / "creds.env", config=tmp_path / "missing.yaml",
        )
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             patch(P_SEND, return_value=mock_result), \
             patch(P_RENDER, side_effect=lambda m, *a: m):
            cmd_send(args)
        assert "Message sent" in capsys.readouterr().out

    def test_exits_without_workspace_url(self, tmp_path):
        args = make_args(
            channel="C1", message=["hi"], workspace=None,
            jitter=0, selection_mode=None,
            env=tmp_path / "creds.env", config=tmp_path / "missing.yaml",
        )
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             pytest.raises(SystemExit) as exc_info:
            cmd_send(args)
        assert exc_info.value.code == 1

    def test_exits_without_message_and_no_config(self, tmp_path):
        args = make_args(
            channel="C1", message=None, workspace="https://test.slack.com",
            jitter=0, selection_mode=None,
            env=tmp_path / "creds.env", config=tmp_path / "missing.yaml",
        )
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             pytest.raises(SystemExit) as exc_info:
            cmd_send(args)
        assert exc_info.value.code == 1

    def test_jitter_calls_sleep(self, tmp_path):
        mock_result = SendResult(ok=True, channel_id="C1", message="hi")
        args = make_args(
            channel="C1", message=["hi"], workspace="https://test.slack.com",
            jitter=5, selection_mode=None,
            env=tmp_path / "creds.env", config=tmp_path / "missing.yaml",
        )
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             patch(P_SEND, return_value=mock_result), \
             patch(P_RENDER, side_effect=lambda m, *a: m), \
             patch(P_SLEEP) as mock_sleep:
            cmd_send(args)
        mock_sleep.assert_called_once()

    def test_zero_jitter_no_sleep(self, tmp_path):
        mock_result = SendResult(ok=True, channel_id="C1", message="hi")
        args = make_args(
            channel="C1", message=["hi"], workspace="https://test.slack.com",
            jitter=0, selection_mode=None,
            env=tmp_path / "creds.env", config=tmp_path / "missing.yaml",
        )
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             patch(P_SEND, return_value=mock_result), \
             patch(P_RENDER, side_effect=lambda m, *a: m), \
             patch(P_SLEEP) as mock_sleep:
            cmd_send(args)
        mock_sleep.assert_not_called()

    def test_failed_send_exits_1(self, tmp_path):
        mock_result = SendResult(ok=False, channel_id="C1", message="hi",
                                 error_code="channel_not_found")
        args = make_args(
            channel="C1", message=["hi"], workspace="https://test.slack.com",
            jitter=0, selection_mode=None,
            env=tmp_path / "creds.env", config=tmp_path / "missing.yaml",
        )
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             patch(P_SEND, return_value=mock_result), \
             patch(P_RENDER, side_effect=lambda m, *a: m), \
             pytest.raises(SystemExit) as exc_info:
            cmd_send(args)
        assert exc_info.value.code == 1

    def test_cycle_mode_with_cli_messages(self, tmp_path, capsys):
        mock_result = SendResult(ok=True, channel_id="C1", message="a", ts="1")
        args = make_args(
            channel="C1", message=["a", "b", "c"],
            workspace="https://test.slack.com",
            jitter=0, selection_mode="cycle",
            env=tmp_path / "creds.env", config=tmp_path / "missing.yaml",
        )
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             patch(P_SEND, return_value=mock_result), \
             patch(P_RENDER, side_effect=lambda m, *a: m), \
             patch(P_PICK, return_value="a") as mock_pick:
            cmd_send(args)
        mock_pick.assert_called_once_with("C1", ["a", "b", "c"], "cycle")
        assert "Message sent" in capsys.readouterr().out

    def test_uses_config_messages_when_no_cli_message(self, tmp_path, capsys):
        channel = ChannelConfig(
            id="C1", name="general",
            messages=["from config"],
            schedules=[ScheduleConfig(cron="0 9 * * *")],
            selection_mode="random",
        )
        config = AppConfig(
            workspace_url="https://test.slack.com",
            channels=[channel],
        )
        config_file = tmp_path / "config.yaml"
        config_file.write_text("placeholder")
        args = make_args(
            channel="C1", message=None, workspace=None,
            jitter=0, selection_mode=None,
            env=tmp_path / "creds.env", config=config_file,
        )
        mock_result = SendResult(ok=True, channel_id="C1", message="from config", ts="1")
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_LOAD_CONFIG, return_value=config), \
             patch(P_VALIDATE), \
             patch(P_SEND, return_value=mock_result), \
             patch(P_RENDER, side_effect=lambda m, *a: m), \
             patch(P_PICK, return_value="from config"):
            cmd_send(args)
        assert "Message sent" in capsys.readouterr().out


# ---------------------------------------------------------------------------
# cmd_status
# ---------------------------------------------------------------------------

class TestCmdStatus:
    def test_calls_print_upcoming_with_count(self):
        args = make_args(config=Path("/fake"), count=3)
        mock_config = MagicMock()
        with patch(P_LOAD_CONFIG, return_value=mock_config), \
             patch(P_PRINT_UPCOMING) as mock_upcoming:
            cmd_status(args)
        mock_upcoming.assert_called_once_with(mock_config, count=3)


# ---------------------------------------------------------------------------
# cmd_validate
# ---------------------------------------------------------------------------

class TestCmdValidate:
    def test_prints_valid_on_success(self, tmp_path, capsys):
        config = AppConfig(workspace_url="https://test.slack.com", channels=[])
        config_file = tmp_path / "config.yaml"
        config_file.write_text('workspace_url: "https://test.slack.com"\nchannels: []\n')
        args = make_args(env=tmp_path / "creds.env", config=config_file)
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_LOAD_CONFIG, return_value=config), \
             patch(P_VALIDATE):
            cmd_validate(args)
        assert "valid" in capsys.readouterr().out

    def test_exits_without_workspace_url(self, tmp_path):
        args = make_args(env=tmp_path / "creds.env", config=tmp_path / "nonexistent.yaml")
        with patch(P_LOAD_CREDS, return_value=MagicMock()), \
             pytest.raises(SystemExit) as exc_info:
            cmd_validate(args)
        assert exc_info.value.code == 1


# ---------------------------------------------------------------------------
# main() integration tests
# ---------------------------------------------------------------------------

class TestMainIntegration:
    def test_no_command_exits_1(self, monkeypatch, capsys):
        monkeypatch.setattr(sys, "argv", ["slack-scheduler"])
        with patch("slack_scheduler.cli.setup_logging"), \
             pytest.raises(SystemExit) as exc_info:
            main()
        assert exc_info.value.code == 1

    def test_token_expired_exits_1(self, monkeypatch, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text('workspace_url: "https://t.com"\nchannels: []\n')
        monkeypatch.setattr(sys, "argv", [
            "slack-scheduler",
            "--config", str(config_file),
            "--env", str(tmp_path / "creds.env"),
            "validate",
        ])
        with patch("slack_scheduler.cli.setup_logging"), \
             patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_LOAD_CONFIG,
                   return_value=AppConfig(workspace_url="https://t.com", channels=[])), \
             patch(P_VALIDATE, side_effect=TokenExpiredError("expired")), \
             pytest.raises(SystemExit) as exc_info:
            main()
        assert exc_info.value.code == 1

    def test_keyboard_interrupt_exits_0(self, monkeypatch, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text('workspace_url: "https://t.com"\nchannels: []\n')
        monkeypatch.setattr(sys, "argv", [
            "slack-scheduler",
            "--config", str(config_file),
            "--env", str(tmp_path / "creds.env"),
            "run",
        ])
        with patch("slack_scheduler.cli.setup_logging"), \
             patch(P_LOAD_CONFIG,
                   return_value=AppConfig(workspace_url="https://t.com", channels=[])), \
             patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             patch(P_RUN_DAEMON, side_effect=KeyboardInterrupt), \
             pytest.raises(SystemExit) as exc_info:
            main()
        assert exc_info.value.code == 0

    def test_slack_api_error_exits_1(self, monkeypatch, tmp_path):
        monkeypatch.setattr(sys, "argv", [
            "slack-scheduler",
            "--config", str(tmp_path / "config.yaml"),
            "--env", str(tmp_path / "creds.env"),
            "send", "--channel", "C1", "--message", "hi",
            "--workspace", "https://test.slack.com",
        ])
        with patch("slack_scheduler.cli.setup_logging"), \
             patch(P_LOAD_CREDS, return_value=MagicMock()), \
             patch(P_VALIDATE), \
             patch(P_SEND, side_effect=SlackAPIError("channel_not_found")), \
             patch(P_RENDER, side_effect=lambda m, *a: m), \
             pytest.raises(SystemExit) as exc_info:
            main()
        assert exc_info.value.code == 1

    def test_value_error_exits_1(self, monkeypatch, tmp_path):
        config_file = tmp_path / "config.yaml"
        config_file.write_text("")  # empty -> ValueError
        monkeypatch.setattr(sys, "argv", [
            "slack-scheduler",
            "--config", str(config_file),
            "--env", str(tmp_path / "creds.env"),
            "status",
        ])
        with patch("slack_scheduler.cli.setup_logging"), \
             pytest.raises(SystemExit) as exc_info:
            main()
        assert exc_info.value.code == 1



import logging
import time
from dataclasses import dataclass

import requests

from slack_scheduler.auth import TokenExpiredError
from slack_scheduler.config import Credentials

log = logging.getLogger(__name__)


class SlackAPIError(Exception):
    def __init__(self, error_code: str):
        self.error_code = error_code
        super().__init__(f"Slack API error: {error_code}")


@dataclass
class SendResult:
    ok: bool
    channel_id: str
    message: str
    ts: str | None = None
    error_code: str | None = None


def send_message(
    channel_id: str,
    message: str,
    credentials: Credentials,
    workspace_url: str,
    dry_run: bool = False,
    max_attempts: int = 3,
    max_rate_limit_retries: int = 5,
) -> SendResult:
    if dry_run:
        log.info(f"[DRY RUN] Would send to {channel_id}: {message!r}")
        return SendResult(ok=True, channel_id=channel_id, message=message)

    attempt = 0
    rate_limit_retries = 0

    while attempt < max_attempts:
        attempt += 1
        try:
            response = _post(channel_id, message, credentials, workspace_url)
            try:
                data = response.json()
            except (ValueError, requests.JSONDecodeError):
                response.raise_for_status()
                raise

            if data.get("ok"):
                return SendResult(
                    ok=True,
                    channel_id=channel_id,
                    message=message,
                    ts=data.get("ts"),
                )

            error = data.get("error", "unknown")

            if error == "invalid_auth":
                raise TokenExpiredError(
                    "Slack token has expired. Update your credentials file."
                )

            if error == "ratelimited":
                rate_limit_retries += 1
                if rate_limit_retries > max_rate_limit_retries:
                    log.error(
                        f"Rate limited {rate_limit_retries} times, giving up."
                    )
                    return SendResult(
                        ok=False,
                        channel_id=channel_id,
                        message=message,
                        error_code="ratelimited",
                    )
                retry_after = int(response.headers.get("Retry-After", 1))
                log.warning(
                    f"Rate limited. Retrying in {retry_after}s "
                    f"(rate limit retry {rate_limit_retries}/{max_rate_limit_retries})."
                )
                time.sleep(retry_after)
                # Don't count rate limits against the attempt budget
                attempt -= 1
                continue

            raise SlackAPIError(error)

        except requests.RequestException as e:
            if attempt == max_attempts:
                raise
            backoff = 2 ** attempt
            log.warning(
                f"Request failed (attempt {attempt}/{max_attempts}): {e}. "
                f"Retrying in {backoff}s."
            )
            time.sleep(backoff)

    raise AssertionError("Unreachable: all attempts should return or raise")


def _post(
    channel_id: str,
    message: str,
    credentials: Credentials,
    workspace_url: str,
) -> requests.Response:
    response = requests.post(
        f"{workspace_url.rstrip('/')}/api/chat.postMessage",
        headers={
            "Authorization": f"Bearer {credentials.xoxc_token}",
            "Content-Type": "application/json; charset=utf-8",
        },
        cookies={"d": credentials.d_cookie},
        json={"channel": channel_id, "text": message},
        timeout=30,
    )
    return response



import logging
from dataclasses import dataclass, field
from datetime import date
from pathlib import Path

import yaml
from dotenv import dotenv_values

log = logging.getLogger(__name__)

VALID_SELECTION_MODES = {"random", "cycle"}


@dataclass
class ScheduleConfig:
    cron: str
    jitter_minutes: int = 0
    skip_weekends: bool = False
    skip_dates: list[str] = field(default_factory=list)


@dataclass
class ChannelConfig:
    id: str
    name: str
    messages: list[str]
    schedules: list[ScheduleConfig]
    selection_mode: str = "random"


@dataclass
class AppConfig:
    workspace_url: str
    channels: list[ChannelConfig]
    default_selection_mode: str = "random"
    skip_dates: list[str] = field(default_factory=list)


@dataclass
class Credentials:
    xoxc_token: str
    d_cookie: str


def _validate_skip_dates(dates: list[str], context: str) -> list[str]:
    """Validate skip_dates entries are valid ISO format dates.

    Args:
        dates: List of date strings to validate.
        context: Description of where these dates come from (e.g., "global skip_dates")
                for error messages.

    Returns:
        The validated list of dates.

    Raises:
        ValueError: If any date is not in YYYY-MM-DD format.
    """
    for d in dates:
        try:
            date.fromisoformat(d)
        except ValueError:
            raise ValueError(
                f"Invalid skip_dates format in {context}: {d!r} "
                "(expected YYYY-MM-DD)"
            )
    return dates


def load_config(config_path: Path) -> AppConfig:
    with open(config_path) as f:
        raw = yaml.safe_load(f)

    if not raw:
        raise ValueError(f"Config file is empty: {config_path}")

    workspace_url = raw.get("workspace_url")
    if not workspace_url:
        raise ValueError("workspace_url is required in config")

    default_mode = raw.get("default_selection_mode", "random")
    if default_mode not in VALID_SELECTION_MODES:
        raise ValueError(
            f"Invalid default_selection_mode: {default_mode!r} "
            f"(expected one of {sorted(VALID_SELECTION_MODES)})"
        )
    global_skip = _validate_skip_dates(raw.get("skip_dates", []), "global skip_dates")

    channels = []
    for idx, ch in enumerate(raw.get("channels", [])):
        if "id" not in ch:
            raise ValueError(
                f"Channel at index {idx} is missing required field 'id'"
            )
        channel_id = ch["id"]
        channel_name = ch.get("name", channel_id)
        schedules = []
        for s_idx, s in enumerate(ch.get("schedules", [])):
            if "cron" not in s:
                raise ValueError(
                    f"Channel '{channel_name}' schedule at index {s_idx} is missing required field 'cron'"
                )
            schedule_skip_dates = s.get("skip_dates", [])
            context = f"channel '{channel_name}' schedule {s_idx} skip_dates"
            validated_skip_dates = _validate_skip_dates(schedule_skip_dates, context)
            schedules.append(ScheduleConfig(
                cron=s["cron"],
                jitter_minutes=s.get("jitter_minutes", 0),
                skip_weekends=s.get("skip_weekends", False),
                skip_dates=validated_skip_dates,
            ))

        messages = ch.get("messages", [])

        if not schedules:
            log.warning(f"Channel {channel_name} ({channel_id}) has no schedules — it will never fire in daemon mode.")
        if not messages:
            log.warning(f"Channel {channel_name} ({channel_id}) has no messages defined.")

        selection_mode = ch.get("selection_mode", default_mode)
        if selection_mode not in VALID_SELECTION_MODES:
            raise ValueError(
                f"Channel '{channel_name}' has invalid selection_mode: {selection_mode!r} "
                f"(expected one of {sorted(VALID_SELECTION_MODES)})"
            )

        channels.append(ChannelConfig(
            id=channel_id,
            name=channel_name,
            messages=messages,
            schedules=schedules,
            selection_mode=selection_mode,
        ))

    return AppConfig(
        workspace_url=workspace_url,
        channels=channels,
        default_selection_mode=default_mode,
        skip_dates=global_skip,
    )


def load_credentials(env_path: Path) -> Credentials:
    if not env_path.exists():
        raise FileNotFoundError(
            f"Credentials file not found: {env_path}\n"
            "Run `slack-scheduler init` to create a template."
        )

    values = dotenv_values(env_path)
    token = values.get("SLACK_XOXC_TOKEN", "")
    cookie = values.get("SLACK_D_COOKIE", "")

    if not token or not cookie:
        raise ValueError(
            f"SLACK_XOXC_TOKEN and SLACK_D_COOKIE must be set in {env_path}. "
            "Run `slack-scheduler init` to create a template, then add your credentials."
        )

    return Credentials(xoxc_token=token, d_cookie=cookie)


def resolve_skip_dates(
    global_dates: list[str], schedule_dates: list[str]
) -> set[date]:
    """Combine global and schedule-specific skip dates into a set of date objects.

    All dates are guaranteed to be valid ISO format strings (validated at config load time).
    """
    combined = set(global_dates) | set(schedule_dates)
    return {date.fromisoformat(d) for d in combined}



"""Tests for slack_scheduler.paths"""
from pathlib import Path
from unittest.mock import patch

from slack_scheduler import paths


class TestPaths:
    def test_config_dir_returns_path(self):
        assert isinstance(paths.config_dir(), Path)

    def test_data_dir_returns_path(self):
        assert isinstance(paths.data_dir(), Path)

    def test_log_dir_is_subdirectory_of_data_dir(self):
        with patch("slack_scheduler.paths.user_data_dir", return_value="/tmp/test-data"):
            assert paths.log_dir() == Path("/tmp/test-data/logs")

    def test_state_dir_is_subdirectory_of_data_dir(self):
        with patch("slack_scheduler.paths.user_data_dir", return_value="/tmp/test-data"):
            assert paths.state_dir() == Path("/tmp/test-data/state")

    def test_config_dir_uses_app_name(self):
        with patch("slack_scheduler.paths.user_config_dir") as mock:
            mock.return_value = "/fake"
            paths.config_dir()
        mock.assert_called_once_with("slack-scheduler")

    def test_data_dir_uses_app_name(self):
        with patch("slack_scheduler.paths.user_data_dir") as mock:
            mock.return_value = "/fake"
            paths.data_dir()
        mock.assert_called_once_with("slack-scheduler")

    def test_log_and_state_dirs_differ(self):
        with patch("slack_scheduler.paths.user_data_dir", return_value="/tmp/test"):
            assert paths.log_dir() != paths.state_dir()



"""Tests for slack_scheduler.selector"""
import json
from pathlib import Path
from unittest.mock import patch

import pytest

from slack_scheduler.selector import _load_state, _save_state, pick_message


# --- pick_message: guard clauses --------------------------------------------

def test_pick_message_raises_on_empty_messages():
    with pytest.raises(ValueError, match="No messages"):
        pick_message("C1", [])


def test_pick_message_single_message_returns_it():
    assert pick_message("C1", ["only one"]) == "only one"


# --- pick_message: random mode ----------------------------------------------

class TestPickMessageRandom:
    def test_random_mode_calls_random_choice(self):
        messages = ["a", "b", "c"]
        with patch("slack_scheduler.selector.random.choice", return_value="b") as mock:
            result = pick_message("C1", messages, mode="random")
        mock.assert_called_once_with(messages)
        assert result == "b"

    def test_default_mode_is_random(self):
        with patch("slack_scheduler.selector.random.choice", return_value="a"):
            assert pick_message("C1", ["a", "b"]) == "a"

    def test_unknown_mode_falls_back_to_random(self):
        with patch("slack_scheduler.selector.random.choice", return_value="x"):
            assert pick_message("C1", ["x", "y"], mode="unknown") == "x"


# --- pick_message: cycle mode -----------------------------------------------

class TestPickMessageCycle:
    def test_cycle_returns_all_messages_before_repeating(self, tmp_path):
        messages = ["a", "b", "c"]
        # Prevent shuffle so order is preserved and deterministic
        with patch("slack_scheduler.selector.random.shuffle"):
            results = [
                pick_message("C1", messages, mode="cycle", state_dir=tmp_path)
                for _ in range(3)
            ]
        # Each message exactly once, in original order (shuffle is no-op)
        assert results == messages

    def test_cycle_reshuffles_after_exhaustion(self, tmp_path):
        messages = ["a", "b"]
        with patch("slack_scheduler.selector.random.shuffle"):
            # Exhaust the cycle
            pick_message("C1", messages, mode="cycle", state_dir=tmp_path)
            pick_message("C1", messages, mode="cycle", state_dir=tmp_path)
            # This should trigger a reshuffle
            pick_message("C1", messages, mode="cycle", state_dir=tmp_path)
        state = json.loads((tmp_path / "C1.json").read_text())
        assert state["index"] == 1

    def test_cycle_uses_channel_specific_state_file(self, tmp_path):
        pick_message("C111", ["x", "y"], mode="cycle", state_dir=tmp_path)
        pick_message("C222", ["a", "b"], mode="cycle", state_dir=tmp_path)
        assert (tmp_path / "C111.json").exists()
        assert (tmp_path / "C222.json").exists()

    def test_cycle_reshuffles_when_messages_change(self, tmp_path):
        original = ["a", "b"]
        modified = ["a", "b", "c"]
        with patch("slack_scheduler.selector.random.shuffle"):
            pick_message("C1", original, mode="cycle", state_dir=tmp_path)
            pick_message("C1", modified, mode="cycle", state_dir=tmp_path)
        state = json.loads((tmp_path / "C1.json").read_text())
        assert sorted(state["shuffled"]) == sorted(modified)

    def test_cycle_creates_state_dir_if_missing(self, tmp_path):
        nested = tmp_path / "deep" / "nested"
        pick_message("C1", ["a", "b"], mode="cycle", state_dir=nested)
        assert nested.is_dir()

    def test_cycle_handles_corrupt_state_file(self, tmp_path):
        state_file = tmp_path / "C1.json"
        state_file.write_text("not valid json")
        result = pick_message("C1", ["a", "b"], mode="cycle", state_dir=tmp_path)
        assert result in ["a", "b"]

    def test_cycle_handles_state_missing_keys(self, tmp_path):
        state_file = tmp_path / "C1.json"
        state_file.write_text(json.dumps({"wrong_key": 1}))
        result = pick_message("C1", ["a", "b"], mode="cycle", state_dir=tmp_path)
        assert result in ["a", "b"]


# --- _load_state / _save_state ----------------------------------------------

class TestStateIO:
    def test_load_returns_none_for_missing_file(self, tmp_path):
        assert _load_state(tmp_path / "nonexistent.json") is None

    def test_load_returns_none_for_invalid_json(self, tmp_path):
        p = tmp_path / "bad.json"
        p.write_text("{ not json }")
        assert _load_state(p) is None

    def test_save_and_load_roundtrip(self, tmp_path):
        p = tmp_path / "state.json"
        state = {"shuffled": ["a", "b"], "index": 1}
        _save_state(p, state)
        assert _load_state(p) == state

    def test_save_is_atomic_no_leftover_files(self, tmp_path):
        p = tmp_path / "state.json"
        _save_state(p, {"shuffled": ["a"], "index": 0})
        assert len(list(tmp_path.iterdir())) == 1
        assert list(tmp_path.iterdir())[0] == p
